==> groupchat-transport-0.9pre4.tar.gz: groupchat-transport-0.9pre4/src/parser.c <==
#include <groupchat.h>

void gc_register_get(jpacket p)
{
    xmlnode_put_attrib(p->x,"from",jid_full(p->to));
    xmlnode_put_attrib(p->x,"to",jid_full(p->from));
    xmlnode_put_attrib(p->x,"type","result");
    xmlnode_insert_tag(p->iq,"name");
    xmlnode_insert_tag(p->iq,"username");
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq,"instructions"),"You should know what you're doing!  Put the name of the private chatroom in the 'name' field and your registered nickname in the 'username' field.",-1);
    ehandler_send(gc__etherx,p->x,p->from->server);
}

void gc_register_set(jpacket p)
{
    xmlnode x;

    jid_set(p->to,xmlnode_get_tag_data(p->iq,"name"),JID_USER);
    jid_set(p->to,xmlnode_get_tag_data(p->iq,"username"),JID_RESOURCE);
    jid_set(p->from,NULL,JID_RESOURCE);

    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x,"type","subscribe");
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),"This will automatically join you to this group every time you are online!",-1);
    xmlnode_put_attrib(x,"to",jid_full(p->from));
    xmlnode_put_attrib(x,"from",jid_full(p->to));
    ehandler_send(gc__etherx,x,p->from->server);
    xmlnode_free(p->x);
}

void gc_presence(gcg g, gcu u, jpacket p)
{
    xmlnode x, body;
    gcu cur;

    switch(jpacket_subtype(p))
    {
        case JPACKET__AVAILABLE:
            /* new group */
            if(g == NULL)
                g = gcg_new(p->to);

            /* new user */
            if(u == NULL)
            {
                if(p->to->resource == NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_NOTALLOWED);
                    return;
                }
                if(gcu_nick(g, p->to->resource) != NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_CONFLICT);
                    return;
                }
                u = gcg_user(g, p->from, p->to->resource);
                for(cur = g->users->next; cur != g->users; cur = cur->next)
                {
                    if(cur == u) continue;
                    x = xmlnode_dup(cur->pres);
                    xmlnode_put_attrib(x,"to",jid_full(u->user));
                    xmlnode_put_attrib(x,"from",jid_full(cur->id));
                    ehandler_send(gc__etherx,x,u->user->server);
                }
            }

            /* nick change */
            if(j_strcmp(p->to->resource, u->nick) != 0)
            {
                if(gcu_nick(g,p->to->resource)!=NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_CONFLICT);
                    return;
                }
                u = gcg_user(g, p->from, p->to->resource);
            }

            xmlnode_free(u->pres);
            u->pres = xmlnode_dup(p->x);

            gcg_send(g, u, p->x);
            break;
        case JPACKET__UNAVAILABLE:
            if(g == NULL || u == NULL)
            {
                xmlnode_free(p->x);
                return;
            }

            /* send the unavailable */
            gcg_send(g, u, p->x);

            gcu_del(g, u);

            /* if the group is empty, ditch it */
            if(g->users->next == g->users)
            {
                g->prev->next = g->next;
                g->next->prev = g->prev;
                pool_free(g->p);
            }
            break;
        default:
            gc_bounce(p, TERROR_NOTIMPL);
    }
}

void gc_breakdance(jpacket p)
{
    gcg g;
    gcu u;
    char *str;

    if(p->from == NULL || p->type == JPACKET_UNKNOWN)
    {
        log_debug(ZONE,"we're dropping a bad packet [%s]",xmlnode2str(p->x));
        xmlnode_free(p->x);
        return;
    }

    /* make sure it's somewhat valid */
    if(p->to == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    if(p->type == JPACKET_IQ)
    {
        if(NSCHECK(p->iq,"jabber:iq:register") && jpacket_subtype(p) == JPACKET__GET)
        {
            gc_register_get(p);
            return;
        }
        if(NSCHECK(p->iq,"jabber:iq:register") && jpacket_subtype(p) == JPACKET__SET)
        {
            gc_register_set(p);
            return;
        }
    }

    /* make sure it's somewhat more valid */
    if(p->to->user == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    /* handle the s10n's right away */
    if(p->type == JPACKET_S10N)
    {
        /* confirm s10n requests, dump rest */
        if(jpacket_subtype(p) == JPACKET__SUBSCRIBE)
        {
            xmlnode_put_attrib(p->x,"to",jid_full(p->from));
            xmlnode_put_attrib(p->x,"from",jid_full(p->to));
            xmlnode_put_attrib(p->x,"type","subscribed");
            ehandler_send(gc__etherx,p->x,p->from->server);
        }else{
            xmlnode_free(p->x);
        }
        return;
    }

    /* find existing group/user */
    g = gcg_get(p->to);
    u = gcu_get(g, p->from);

    /* presence first, might create the group/user */
    if(p->type == JPACKET_PRESENCE)
    {
        gc_presence(g, u, p);
        return;
    }

    /* no group? lost cause */
    if(g == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    /* sender is in the group, and recipient is the group */
    if(u != NULL && (p->to->resource == NULL || strcmp(p->to->resource,u->nick) == 0))
    {
        if(p->type != JPACKET_MESSAGE)
        {
            gc_bounce(p, TERROR_NOTALLOWED);
            return;
        }

        log_debug(ZONE,"it's a normal message to the room");

        if(jpacket_subtype(p) != JPACKET__GROUPCHAT)
            xmlnode_put_attrib(p->x,"type","groupchat");

        /* catch the invite hack */
        str = xmlnode_get_tag_data(p->x,"subject");
        if(str != NULL && strncasecmp(str,"invite:",7) == 0)
        {
            str += 7;
            xmlnode_put_attrib(p->x,"to",str);
            xmlnode_put_attrib(p->x,"from",jid_full(u->id));
            jpacket_reset(p);
            if(p->to == NULL)
            { /* bad invite address, crap, drop it */
                xmlnode_free(p->x);
                return;
            }
            ehandler_send(gc__etherx,p->x,p->to->server);
        }
        gcg_send(g, u, p->x);
        return;
    }

    /* all that's left is private messages */

    /* you have to be in the chatroom to send a private message! */
    if(u == NULL)
    {
        gc_bounce(p, TERROR_FORBIDDEN);
        return;
    }
    xmlnode_put_attrib(p->x, "from", jid_full(u->id));

    /* get the recipient */
    u = gcu_nick(g, p->to->resource);

    /* nobody here w/ that nick */
    if(u == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    log_debug(ZONE,"it's a private message");

    /* private messages can't be groupchat */
    if(jpacket_subtype(p) == JPACKET__GROUPCHAT)
        xmlnode_put_attrib(p->x,"type","chat");

    /* send it away */
    xmlnode_put_attrib(p->x,"to",jid_full(u->user));
    ehandler_send(gc__etherx, p->x, u->user->server);
}

==> groupchat-transport-1.0.0.tar.gz: groupchat-transport-1.0.0/src/parser.c <==
#include <groupchat.h>

void gc_register_get(jpacket p)
{
    xmlnode_put_attrib(p->x,"from",jid_full(p->to));
    xmlnode_put_attrib(p->x,"to",jid_full(p->from));
    xmlnode_put_attrib(p->x,"type","result");
    xmlnode_insert_tag(p->iq,"name");
    xmlnode_insert_tag(p->iq,"username");
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq,"instructions"),"You should know what you're doing!  Put the name of the private chatroom in the 'name' field and your registered nickname in the 'username' field.",-1);
    ehandler_send(gc__etherx,p->x,p->from->server);
}

void gc_register_set(jpacket p)
{
    xmlnode x;

    jid_set(p->to,xmlnode_get_tag_data(p->iq,"name"),JID_USER);
    jid_set(p->to,xmlnode_get_tag_data(p->iq,"username"),JID_RESOURCE);
    jid_set(p->from,NULL,JID_RESOURCE);

    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x,"type","subscribe");
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),"This will automatically join you to this group every time you are online!",-1);
    xmlnode_put_attrib(x,"to",jid_full(p->from));
    xmlnode_put_attrib(x,"from",jid_full(p->to));
    ehandler_send(gc__etherx,x,p->from->server);
    xmlnode_free(p->x);
}

void gc_presence(gcg g, gcu u, jpacket p)
{
    xmlnode x;
    gcu cur;

    switch(jpacket_subtype(p))
    {
        case JPACKET__AVAILABLE:
            /* new group */
            if(g == NULL)
                g = gcg_new(p->to);

            /* new user */
            if(u == NULL)
            {
                if(p->to->resource == NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_NOTALLOWED);
                    return;
                }
                if(gcu_nick(g, p->to->resource) != NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_CONFLICT);
                    return;
                }
                u = gcg_user(g, p->from, p->to->resource);
                for(cur = g->users->next; cur != g->users; cur = cur->next)
                {
                    if(cur == u) continue;
                    x = xmlnode_dup(cur->pres);
                    xmlnode_put_attrib(x,"to",jid_full(u->user));
                    xmlnode_put_attrib(x,"from",jid_full(cur->id));
                    ehandler_send(gc__etherx,x,u->user->server);
                }
            }

            /* nick change */
            if(j_strcmp(p->to->resource, u->nick) != 0)
            {
                if(gcu_nick(g,p->to->resource)!=NULL)
                {
                    x = xmlnode_new_tag("message");
                    xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
                    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(p->x,"to"));
                    xmlnode_free(p->x);
                    p = jpacket_new(x);
                    gc_bounce(p, TERROR_CONFLICT);
                    return;
                }
                u = gcg_user(g, p->from, p->to->resource);
            }

            xmlnode_free(u->pres);
            u->pres = xmlnode_dup(p->x);

            gcg_send(g, u, p->x);
            break;
        case JPACKET__UNAVAILABLE:
            if(g == NULL || u == NULL)
            {
                xmlnode_free(p->x);
                return;
            }

            /* send the unavailable */
            gcg_send(g, u, p->x);

            gcu_del(g, u);

            /* if the group is empty, ditch it */
            if(g->users->next == g->users)
            {
                message prev;
                g->prev->next = g->next;
                g->next->prev = g->prev;
                for(prev=g->history;
                    g->history!=NULL;
                    g->history=g->history->next,pool_free(prev->p),prev=g->history);
                pool_free(g->p);
            }
            break;
        default:
            gc_bounce(p, TERROR_NOTIMPL);
    }
}

void gc_breakdance(jpacket p)
{
    gcg g;
    gcu u;
    char *str;

    if(p->from == NULL || p->type == JPACKET_UNKNOWN)
    {
        log_debug(ZONE,"we're dropping a bad packet [%s]",xmlnode2str(p->x));
        xmlnode_free(p->x);
        return;
    }

    /* make sure it's somewhat valid */
    if(p->to == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    if(p->type == JPACKET_IQ)
    {
        if(NSCHECK(p->iq,"jabber:iq:register") && jpacket_subtype(p) == JPACKET__GET)
        {
            gc_register_get(p);
            return;
        }
        if(NSCHECK(p->iq,"jabber:iq:register") && jpacket_subtype(p) == JPACKET__SET)
        {
            gc_register_set(p);
            return;
        }
    }

    /* make sure it's somewhat more valid */
    if(p->to->user == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    /* handle the s10n's right away */
    if(p->type == JPACKET_S10N)
    {
        /* confirm s10n requests, dump rest */
        if(jpacket_subtype(p) == JPACKET__SUBSCRIBE)
        {
            xmlnode_put_attrib(p->x,"to",jid_full(p->from));
            xmlnode_put_attrib(p->x,"from",jid_full(p->to));
            xmlnode_put_attrib(p->x,"type","subscribed");
            ehandler_send(gc__etherx,p->x,p->from->server);
        }else{
            xmlnode_free(p->x);
        }
        return;
    }

    /* find existing group/user */
    g = gcg_get(p->to);
    u = gcu_get(g, p->from);

    /* presence first, might create the group/user */
    if(p->type == JPACKET_PRESENCE)
    {
        gc_presence(g, u, p);
        return;
    }

    /* no group? lost cause */
    if(g == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    /* sender is in the group, and recipient is the group */
    if(u != NULL && (p->to->resource == NULL ))
    {
        if(p->type != JPACKET_MESSAGE)
        {
            gc_bounce(p, TERROR_NOTALLOWED);
            return;
        }

        log_debug(ZONE,"it's a normal message to the room");

        if(jpacket_subtype(p) != JPACKET__GROUPCHAT)
            xmlnode_put_attrib(p->x,"type","groupchat");

        /* catch the invite hack */
        str = xmlnode_get_tag_data(p->x,"subject");
        if(str != NULL && strncasecmp(str,"invite:",7) == 0)
        {
            str += 7;
            xmlnode_put_attrib(p->x,"to",str);
            xmlnode_put_attrib(p->x,"from",jid_full(u->id));
            jpacket_reset(p);
            if(p->to == NULL)
            { /* bad invite address, crap, drop it */
                xmlnode_free(p->x);
                return;
            }
            ehandler_send(gc__etherx,p->x,p->to->server);
        }
        gcg_send(g, u, p->x);
        return;
    }

    /* all that's left is private messages */

    /* you have to be in the chatroom to send a private message! */
    if(u == NULL)
    {
        gc_bounce(p, TERROR_FORBIDDEN);
        return;
    }
    xmlnode_put_attrib(p->x, "from", jid_full(u->id));

    /* get the recipient */
    u = gcu_nick(g, p->to->resource);

    /* nobody here w/ that nick */
    if(u == NULL)
    {
        gc_bounce(p, TERROR_NOTFOUND);
        return;
    }

    log_debug(ZONE,"it's a private message");

    /* private messages can't be groupchat */
    if(jpacket_subtype(p) == JPACKET__GROUPCHAT)
        xmlnode_put_attrib(p->x,"type","chat");

    /* send it away */
    xmlnode_put_attrib(p->x,"to",jid_full(u->user));
    ehandler_send(gc__etherx, p->x, u->user->server);
}
